function [Ms,hred_array,hrem_array] = deterministic_mr1l_approx_get_lattice_sizes(I,r1l_z_array_cell,r1l_M)
%DETERMINISTIC_MR1L_GET_LATTICE_SIZES Computes multiple rank-1 lattices by Algorithm 2.1.

if size(I,1) == 1
  Ms = 2;
  hred_array = 1;
  hrem_array = 0;
  return;
end

I = int64(I);
if iscell(r1l_z_array_cell)
  for t = 1:length(r1l_z_array_cell)
    r1l_z_array_cell{t} = int64(r1l_z_array_cell{t});
  end
else
  r1l_z_array_cell = int64(r1l_z_array_cell);
end
N_I = max(max(I,[],1)-min(I,[],1)+1);
[s,d] = size(I);

r1l_z_output = [];

if ~iscell(r1l_z_array_cell) && (d * double(N_I) * double(r1l_M) < 2^63)
  freq_times_z_precomputed = I(:,1) * r1l_z_array_cell(1);
  for t = 2:d
    freq_times_z_precomputed = freq_times_z_precomputed + I(:,t) * r1l_z_array_cell(t);
  end
else
  freq_times_z_precomputed = [];
end

% htimesz = sum(freq .* r1l_z, 2);
% htimesz = freq(:,1) * r1l_z(1);
% for t = 2:d
%   if double(max(freq(:,t))) > 2^63 / d / double(r1l_z(t))
%     error('quantities too large');
%   end
%   htimesz = htimesz + freq(:,t) * r1l_z(t);
% end
% 
% if length(unique(mod(htimesz,r1l_M))) ~= size(freq,1)
%   error('rank-1 lattice is not reconstructing');
% end

P_upper_bound = floor( 2.832*s * log(d*double(N_I)*double(r1l_M))/log(s) * log(2.3*s*log(d*double(N_I)*double(r1l_M))/log(s)) );
% max_P = min(P_upper_bound,2.0*r1l_M);
max_P = P_upper_bound;
if max_P > 2e9
  warning('Upper bound on P too large => restricting to 2e9. Search may fail!');
  max_P = 2e9;
end
max_P = int64(max_P);
primes_array = primes(max_P);
primes_array = primes_array(primes_array >= size(I,1));

Ms = [];
hred_array = [];
hrem_array = [];
iP = 0;
% tilde_I=int64(zeros(0,size(I,2)));
ind_tilde_I=[];
while size(I,1) > length(ind_tilde_I)
  iP = iP + 1;
  if iP > length(primes_array)
    error('lattice not found');
  end
  
  P = int64(primes_array(iP));

  if ~isempty(freq_times_z_precomputed)
    freq_times_z_mod_P = mod(freq_times_z_precomputed,P);
    r1l_z = mod(r1l_z_array_cell,P);
  else
    if iscell(r1l_z_array_cell)
      r1l_z = int64(ones(1,d));
      for t = 1:d
        for ii = 1:length(r1l_z_array_cell{t})
          r1l_z(t) = mod(r1l_z(t) * mod(r1l_z_array_cell{t}(ii),P), P);
        end
      end
    else
      r1l_z = mod(r1l_z_array_cell,P);
    end
    freq_times_z = mod(mod(I(:,1),P) * r1l_z(1),P);
    for t = 2:d
      freq_times_z = freq_times_z + mod(mod(I(:,t),P) * r1l_z(t),P);
    end
    freq_times_z_mod_P = mod(freq_times_z,P);
  end

  [tmpsort,ind1]=sort(freq_times_z_mod_P);
  tmpsortdiff=tmpsort(2:end)-tmpsort(1:end-1);
  ind2=[];
  if(tmpsortdiff(1)~=0)
    ind2=[ind2;1];
  end
  if(tmpsortdiff(end)~=0)
    ind2=[ind2;length(tmpsort)];
  end
  ind2tmp=find(tmpsortdiff(1:end-1).*tmpsortdiff(2:end)~=0);
  ind2=[ind2;ind2tmp+1];
  ind_tilde_I_new = unique([ind_tilde_I;ind1(ind2)]);
%   tilde_I_new=unique([tilde_I;I(ind1(ind2),:)],'rows');
  hred = length(ind_tilde_I_new)-length(ind_tilde_I);
  hrem = size(I,1) - length(ind_tilde_I);
  if hred < hrem / 2
    continue
  end
%   tilde_I = tilde_I_new;
  ind_tilde_I = ind_tilde_I_new;

  Ms = [Ms, P];
  hred_array = [hred_array; hred];
  hrem_array = [hrem_array; size(I,1) - length(ind_tilde_I)];
  r1l_z_output = [r1l_z_output; r1l_z];
  iP = 0;
end

Ms = double(Ms);
r1l_z_output = double(r1l_z_output);

freq = double(I);
freq_reco = zeros(0, size(I,2));
for ell = 1:length(Ms)
  freq_reco_ell = get_reconstructable_freq(freq,r1l_z_output(ell,:),Ms(ell));
  freq_reco = unique([freq_reco; freq_reco_ell], 'rows');
end

if size(I,1) ~= size(freq_reco)
  error('final check failed');
end

end

